local FS = LibStub("AceAddon-3.0"):GetAddon("FaceShooter")

----------
--- FS ---
----------

-- Copies the src table while reusing the memory already allocated for dst table and skipping (static) functions
function FS:CopyTable(src, dst)
	
	-- clear the destination table so therere no set variables that are nil in the source table
	for akey, avalue in pairs(dst) do
		if type(avalue) == "table" then
			for bkey, bvalue in pairs(avalue) do
				if type(bvalue) == "table" then
					for ckey, cvalue in pairs(bvalue) do
						if type(cvalue) == "table" then
							for dkey, dvalue in pairs(cvalue) do
								if type(dvalue) ~= "function" then dst[akey][bkey][ckey][dkey] = nil
								end
							end
						elseif type(cvalue) ~= "function" then dst[akey][bkey][ckey] = nil
						end
					end
				elseif type(bvalue) ~= "function" then dst[akey][bkey] = nil
				end
			end
		elseif type(avalue) ~= "function" then dst[akey] = nil
		end
	end

	
	-- copy table by hand to avoid referencing and memory issues
	for akey, avalue in pairs(src) do
		if type(avalue) == "table" then
			for bkey, bvalue in pairs(avalue) do
				if type(bvalue) == "table" then
					for ckey, cvalue in pairs(bvalue) do
						if type(cvalue) == "table" then
							for dkey, dvalue in pairs(cvalue) do
								if type(dvalue) ~= "function" then dst[akey][bkey][ckey][dkey] = dvalue
								end
							end
						elseif type(cvalue) ~= "function" then dst[akey][bkey][ckey] = cvalue
						end
					end
				elseif type(bvalue) ~= "function" then dst[akey][bkey] = bvalue
				end
			end
		elseif type(avalue) ~= "function" then dst[akey] = avalue
		end
	end
end

-- Gets a table from the pool or creates a new one
function FS:GetTable()
	return table.remove(self.freeTables) or {}
end

-- Empties the given table and puts it in the pool
function FS:RecycleTable(tbl)
	for key in pairs(tbl) do
		tbl[key] = nil
	end
	table.insert(self.freeTables, tbl)
end

-- removes an existing shot object
function FS:RemoveButton(name)
	self[name]["mov"]:Hide()
	self:RecycleTable(self[name]["mov"])
	self[name]["mov"] = nil
	self[name]["btn"]:Hide()
	self:RecycleTable(self[name]["btn"])
	self[name]["btn"] = nil
	self[name]["cdf"]:Hide()
	self:RecycleTable(self[name]["cdf"])
	self[name]["cdf"] = nil
	self[name].resources = nil
	self:RecycleTable(self[name])
	self[name] = nil
end

-- simple function to round numbers
function FS:Round(num, digits)
	if num == nil then return 0 end
	if not tonumber(num) then return num end
	local mult = 10 ^ (digits or 0)
	return floor(num * mult + 0.5) / mult
end

-- called when a valid new target is acquired
function FS:AcquireTarget()
	self.tarGUID = UnitGUID("target")
	if self.state == "idle" then self.state = "waiting" end
	for i = 1, 2 do
		if self.db.profile["button"..i]["enabled"] then self["button"..i]:Show() end
	end
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.AcquireTarget then module:AcquireTarget() end
	end
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("UNIT_SPELLCAST_START")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self.resources:GetDebuffs()
	self:Scheduler()
end

-- called when target is lost or not valid
function FS:LooseTarget()
	self.tarGUID = nil
	self.state = "idle"
	for i = 1, 2 do
		if self.db.profile["button"..i]["enabled"] then self["button"..i]:Hide() end
	end
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.LooseTarget then module:LooseTarget() end
	end
	self:UnregisterEvent("UNIT_AURA")
	self:UnregisterEvent("UNIT_SPELLCAST_SENT")
	self:UnregisterEvent("UNIT_SPELLCAST_START")
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

-- calculate the average of a given statType over a given amount of datapoints
function FS:GetAverage(statType, data, dataPoints)
	dataPoints = dataPoints or 10
	if not self[statType] then self[statType] = {} end
	table.insert(self[statType], 1, data)
	table.remove(self[statType], dataPoints)
	local sum, num = 0, 0
	for key, val in pairs(self[statType]) do
		sum = sum + val
		num = num + 1
	end
	return (sum / num)
end

-- determine which options need to be activated or deactivated
function FS:CheckForDisabled()
	if self.db.profile.sps.suggestSpS == "never" then FS.options.args.sps.args.saveFocusSpS.disabled = true
	else FS.options.args.sps.args.saveFocusSpS.disabled = false
	end
	if self.db.profile.ba.suggestBA == "never" then FS.options.args.ba.args.saveFocusBA.disabled = true
	else FS.options.args.ba.args.saveFocusBA.disabled = false
	end
	if self.db.profile.ba.useTraps == "never" or self.db.profile.ba.suggestET == "never" then FS.options.args.ba.args.saveFocusET.disabled = true
	else FS.options.args.ba.args.saveFocusET.disabled = false
	end
end

-- check for interrupts
function FS:CheckForInterrups()
	local spellName, rank, displayName, icon, startTime, endTime = UnitCastingInfo("player")
	if spellName then self.castEndsAt = endTime / 1000 end
	-- spellNum table should be empty while not casting anything
	if self.state ~= "castStarted" then
		if next(self.spellNum) then
			for key, val in pairs(self.spellNum) do
				if val == self.spellName[0] then
					self.spellName[0] = nil
					self.spellSentAt[0] = nil
					self.spellStartedAt[0] = nil
					self.spellEndedAt[0] = nil
				end
				if val == GetSpellInfo(self.spellID["SS"]) then self.resources.focusGains["SS"] = self.resources.focusGains["SS"] - 1
				elseif val == GetSpellInfo(self.spellID["CoS"]) then self.resources.focusGains["CoS"] = self.resources.focusGains["CoS"] - 1
				elseif val == GetSpellInfo(self.spellID["AimS"]) then self.resources.focusGains["AimS"] = self.resources.focusGains["AimS"] - 1
				end
				self.spellNum[key] = nil
			end
			if self.resources.focusGains["SS"] < 0 then self.resources.focusGains["SS"] = 0 end
			if self.resources.focusGains["CoS"] < 0 then self.resources.focusGains["CoS"] = 0 end
			if self.resources.focusGains["AimS"] < 0 then self.resources.focusGains["AimS"] = 0 end
		end
	
	-- if in castStarted state but no spell casting
	elseif not spellName then
		local now = GetTime()
		if now < self.castEndsAt or now > (self.castEndsAt + self.avrgLatency) then
			for key, val in pairs(self.spellNum) do	self.spellNum[key] = nil end
			self.spellName[0] = nil
			self.spellName[1] = nil
			self.spellSentAt[0] = nil
			self.spellSentAt[1] = nil
			self.spellStartedAt[0] = nil
			self.spellStartedAt[1] = nil
			self.spellEndedAt[0] = nil
			self.spellEndedAt[1] = nil
			self.resources.focusGains["SS"] = 0
			self.resources.focusGains["CoS"] = 0
			self.resources.focusGains["AimS"] = 0
			self.state = "waiting"
		end
	end
end

-- checks if player is in a raid
function FS:IsInRaid()
	return UnitInRaid("player")
end

-- checks if player is in a party
function FS:IsInParty()
	if GetNumPartyMembers() > 0 then
		return true
	else
		return false
	end
end

-- checks if player is solo
function FS:IsSolo()
	if GetNumPartyMembers() > 0 or UnitInRaid("player") then
		return false
	else
		return true
	end
end

-- checks if player is mounted or in a vehicle
function FS:IsMounted()
	if IsMounted() or UnitInVehicle("player") then
		return true
	else
		return false
	end
end

-- checks if target is a valid target according to settings
function FS:IsLegitTarget()
	-- if player has no target, target is friendly, target is dead or a player character
	if UnitName("target") == nil or UnitIsFriend("player","target") ~= nil or UnitHealth("target") == 0 or UnitIsPlayer("target") then
		return false
	-- if player is in raid and showInRaid is false or player in party and showInParty is false etc.
	elseif (self:IsInRaid() and not self.db.profile.showInRaid) or (self:IsInParty() and not self.db.profile.showInParty) or (self:IsSolo() and not self.db.profile.showWhenSolo) then
		return false
	-- if player is flagged as mounted
	elseif self.isMounted then
		return false
	else
		return true
	end
end

-- reads the talents and returns the spec
function FS:GetTalents()
	-- Find out which talents are available to the player
	self.talents = {}
	local tab
	for tab = 1, 3 do
		local numTalents = GetNumTalents(tab)
		local talent = 0
		for talent = 1, numTalents do
			local nameTalent, _, _, _, currentRank, _, _, _ = GetTalentInfo(tab, talent)
			if nameTalent then
				for key, ID in pairs(self.spellID) do
					if nameTalent == GetSpellInfo(ID) then self.talents[ID] = currentRank end
				end
			end
		end
	end

	if self.talents[self.spellID["BM"]] > 0 then return "BM"
	elseif self.talents[self.spellID["CS"]] > 0 then return "MM"
	elseif self.talents[self.spellID["BA"]] > 0 then return "SV"
	else return "none"
	end
end

-- calculate the spellcosts considering talents and glyphs
function FS:GetSpellCosts()
	self.cost["ES"] = (50 - FS.talents[FS.spellID["EFF"]] * 2)
	self.cost["KC"] = (40 - FS.glyphs[FS.glyphID["KC"]] * 3)
	self.cost["CS"] = (50 - FS.talents[FS.spellID["EFF"]] * 2)
	self.cost["BA"] = 35
	self.cost["SpS"] = 25
	self.cost["AS"] = (25 - FS.talents[FS.spellID["EFF"]])
	self.cost["AimS"] = 50
	self.cost["TL"] = (20 - FS.glyphs[FS.glyphID["TL"]] * 10)
end

-- reads the gylphs
function FS:GetGlyphs()
	-- Find out which glyphs are available to the player
	if not self.glyphs then
		self.glyphs = {}
	else
		for key in pairs(self.glyphs) do self.glyphs[key] = nil end
	end
	
	local slot
	for slot = 1, 9 do
		local _, _, _, glyphID = GetGlyphSocketInfo(slot, GetActiveTalentGroup())
		if glyphID then self.glyphs[glyphID] = 1 end
	end
	for key, ID in pairs(self.glyphID) do
		if not self.glyphs[ID] then self.glyphs[ID] = 0 end
	end
end

-- read the gear names and sets set boni variables accordingly
function FS:GetItems()
	if self.sets then self:RecycleTable(self.sets) end
	self.sets = self:GetTable()
	local tooltip = CreateFrame("GameTooltip", "tooltip")
	tooltip:SetOwner(WorldFrame, "ANCHOR_NONE")
	tooltip:AddFontStrings(tooltip:CreateFontString("$parentTextLeft1", nil, "GameTooltipText"), tooltip:CreateFontString("$parentTextRight1", nil, "GameTooltipText"))
	-- Slots 1 to 10 contain all items present in current T Sets - expand to 0 to 18 if necessary
	for slot = 1, 10 do
		local hasItem = tooltip:SetInventoryItem("player", slot)
		if hasItem then
			name = getglobal("tooltipTextLeft1"):GetText()
			local _, itemLink = GetItemInfo(name)
			if itemLink then
				local _, _, _, _, equippedItemID = string.find(itemLink, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+)")
				equippedItemID = tonumber(equippedItemID)
				-- Add up all found setitems
				for setName, set in pairs(self.itemID) do
					for setItem, setItemID in pairs(set) do
						if equippedItemID == setItemID then
							if not self.sets[setName] then
								self.sets[setName] = 1
							else
								self.sets[setName] = self.sets[setName] + 1
							end
						end
					end
				end
			end
		end
	end
end

-- get the current gcd and return start and duration
function FS:GetGCD()
	local start, duration = GetSpellCooldown(FS.spellID["SS"])
	local gcd = start + duration
	local now = GetTime() 
	-- if cast has just been sent and the gcd is set but lower than 0.5 its propably still the last casts GCD
	if self.state == "castSent" and gcd > 0 and gcd < now + 0.5 then
		self.gcd = now + 1
		return now, 1
	-- anytime else just use the values of GetSpellCooldown
	else
		self.gcd = gcd
		return start, duration
	end
end

-- returns how long the given cast will take
function FS:GetCastDelta(spell, haste, now, mmm)
	if spell == GetSpellInfo(self.spellID["SS"]) or spell == GetSpellInfo(self.spellID["CoS"]) then
		delta = 2 / haste
		if not self.sets then FS:GetItems() end
		if self.sets.t11 and self.sets.t11 >= 4 then delta = delta - 0.2 end
		if delta < 1 then delta = 1 end
	elseif spell == GetSpellInfo(self.spellID["AimS"]) or spell == string.sub(GetSpellInfo(self.spellID["AimS"]), 1, -2) then
		if mmm > now then
			delta = 1
		else
			delta = 2.9 / haste
			if delta < 1 then delta = 1	end
		end
	else
		delta = 1
	end
	return delta
end

----------------
-- resources ---
----------------

-- gets the resources as they are at the moment
local function GetResources(self)
	local now = GetTime()
	local focus = UnitPower("player")
	local rangedAttackSpeed = UnitRangedDamage("player")
	local start, duration = GetSpellCooldown(FS.spellID["SS"])
	local gcd = start + duration
	
	-- copy the current spellqueues
	FS:CopyTable(FS.spellName, self.spellName)
	FS:CopyTable(FS.spellSentAt, self.spellSentAt)
	FS:CopyTable(FS.spellStartedAt, self.spellStartedAt)
	FS:CopyTable(FS.spellEndedAt, self.spellEndedAt)
	
	-- get haste, targets hp, focus and time
	self.focus = focus
	self.now = now
	self.tarHP = UnitHealth("target") / UnitHealthMax("target")
	self.haste = 1 + GetRangedHaste() / 100
	-- Neither Improved Steady Shot nor the 10% Raid Physical Haste buff affect focus regen
	self.relevantHaste = self.haste
	if self.buffs["ISS"] > self.now then self.relevantHaste = self.relevantHaste / (1 + 0.05 * FS.talents[FS.spellID["ISS"]]) end
	if self.buffs["HP"] > self.now or self.buffs["IIT"] > self.now or self.buffs["WFT"] > self.now then self.relevantHaste = self.relevantHaste / 1.1 end
	-- T13 2pc Set bonus doubles focus gain due to SS and CoS
	if not FS.sets then FS:GetItems() end
	if FS.sets.T13 and FS.sets.T13 >= 2 then
		if self.tarHP <= 0.25 then self.gains = (9 + FS.talents[FS.spellID["TER"]] * 3) * 2
		else self.gains = 18
		end
	else
		if self.tarHP <= 0.25 then self.gains = 9 + FS.talents[FS.spellID["TER"]] * 3
		else self.gains = 9
		end
	end
	
	-- get the CDs of the various used spells
	start, duration = GetSpellCooldown(FS.spellID["KS"])
	if (start + duration) == gcd then self.cooldowns["KS"] = 0
	else self.cooldowns["KS"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["CS"])
	if (start + duration) == gcd then self.cooldowns["CS"] = 0
	else self.cooldowns["CS"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["BA"])
	if (start + duration) == gcd then self.cooldowns["BA"] = 0
	else self.cooldowns["BA"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["ES"])
	if (start + duration) == gcd then self.cooldowns["ES"] = 0
	else self.cooldowns["ES"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["AimS"])
	if (start + duration) == gcd then self.cooldowns["AimS"] = 0
	else self.cooldowns["AimS"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["KC"])
	if (start + duration) == gcd then self.cooldowns["KC"] = 0
	else self.cooldowns["KC"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["ET"])
	if (start + duration) == gcd then self.cooldowns["ET"] = 0
	else self.cooldowns["ET"] = start + duration
	end
	
	-- get buffs and debuffs
	self:GetBuffs()
	self:GetDebuffs()
end

-- sets the resources to the given values
local function SetResources(self, resources)
	FS:CopyTable(resources, self)
end

-- account for focus gains that happen after the cast
local function FocusGains(self)
	for key, val in pairs(self.focusGains) do
		if (key == "SS" or key == "CoS") and self.tarHP <= 0.25 then 
			self.focus = self.focus + self.gains * val
			self.focusGains[key] = 0
		elseif key == "RiF" then
			self.focus = self.focus + 40 * val
			self.focusGains["RiF"] = 0
		elseif key == "AimS" then
			self.focus = self.focus - 50 * val
			self.focusGains["AimS"] = 0
		end
	end
	if FS.talents[FS.spellID["KinS"]] > 0 and self.focus > (100  + FS.talents[FS.spellID["KinS"]] * 5) then self.focus = 100 + FS.talents[FS.spellID["KinS"]] * 5
	elseif FS.talents[FS.spellID["KinS"]] == 0 and self.focus > 100 then self.focus = 100
	end
end

-- returns the amount of casts of the given spells since the given point in time
local function NumCasts(self, since, ...)
	local numCasts, numArgs, cmpr = 0, select("#", ...)
	for k = 1, numArgs do
		cmpr = select(k, ...)
		for i = 0, 4 do
			if self.spellName[i] == cmpr and self.spellSentAt[i] >= since then numCasts = numCasts + 1 end
		end
	end
	return numCasts
end

-- get the buffs on yourself
local function GetBuffs(self)
	local now = GetTime()
	for buff, val in pairs(self.buffs) do
		local spellName = GetSpellInfo(FS.spellID[buff])
		local name, _, _, cnt, _, dur, expTime, caster = UnitBuff("player", spellName)
		if name then
			self.buffs[buff] = expTime
		else
			if buff == "LnL" then
				local explo, arcs = GetSpellInfo(FS.spellID["ES"]), GetSpellInfo(FS.spellID["AS"])
				local num = self:NumCasts((self.buffs["LnL"] - 12), explo, arcs)
				if num > 2 or (num == 2 and not self:IsLastShot(explo) and not self:IsLastShot(arcs)) then self.buffs["LnL"] = 0 end
			elseif buff == "MMM" then
				local aims = GetSpellInfo(FS.spellID["AimS"])
				local num = self:NumCasts((self.buffs["MMM"] - 12), aims)
				if num > 1 or (num == 1 and not self:IsLastShot(aims)) then self.buffs["MMM"] = 0 end
			else
				self.buffs[buff] = 0
			end
		end
		-- check if T13p4 proc has been registered
		if self.buffs["CHR"] > now and (now - FS.lastT13procAt) > 20 then FS.lastT13procAt = now end
	end
end

-- get the debuffs on target
local function GetDebuffs(self)
	local now, filter = GetTime()
	for debuff, val in pairs(self.debuffs) do
		if debuff == "SpS" then	filter = "PLAYER" else filter = nil end
		local name, _, _, cnt, _, dur, expTime, caster = UnitDebuff("target", GetSpellInfo(FS.spellID[debuff]), nil, filter)
		-- if spell has been found use the found duration
		if name then
			-- if its HM we dont mind if the player casted it himself
			if name == GetSpellInfo(FS.spellID["HM"]) or name == GetSpellInfo(FS.spellID["MfD"]) then
				self.debuffs[debuff][FS.tarGUID] = expTime
			-- otherwise only consider our own debuffs
			elseif caster == "player" then
				self.debuffs[debuff][FS.tarGUID] = expTime
			else
				self.debuffs[debuff][FS.tarGUID] = 0
			end
		-- if spell with that name hast just been cast assume it just hasnt reached the server yet
		elseif self.lastCast[debuff] and now < (self.lastCast[debuff] + 1 + FS.avrgLatency) then
			self.debuffs[debuff][FS.tarGUID] = now + 15
		-- otherwise set debuff duration to 0
		else
			self.debuffs[debuff][FS.tarGUID] = 0
		end
	end
end

-- when ooc purge remaining debuffs on everything thats not your target
local function PurgeDebuffs(self)
	for debuff, val in pairs(self.debuffs) do
		for id, expTime in pairs(val) do
			if id ~= FS.tarGUID then self.debuffs[debuff][id] = nil end
		end
	end
end

-- returns assumed changes to resources when casting given spell
local function CastSpell(self, parent)
	local now = GetTime()

	-- get the self.spell for which were trying to find the outcome
	if self.spellName[0] then self.spell = self.spellName[0]
	elseif self.spellName[1] then self.spell = self.spellName[1]
	end

	-- calculate how long we have to wait after casting given spell
	local delta = FS:GetCastDelta(self.spell, self.haste, self.buffs["MMM"], self.now)
	
	-- if a cast has been started use that starting time as base for the timer calculations
	if parent ~= "fs" then self.now = self.spellSentAt[1]
	elseif FS.state ~= "waiting" and self.spellName[0] then self.now = self.spellSentAt[0]
	elseif FS.state ~= "waiting" and self.spellName[1] then self.now = self.spellSentAt[1]
	elseif FS.state ~= "waiting" and self.spellName[2] then self.now = self.spellSentAt[2]
	else self.now = now
	end
	
	-- Calculate cooldowns, buffs and debuffs
	self:CalcTimers(parent)
	
	-- Calculate focus - not required after focus has been subtracted serverside
	if parent ~= "fs" then self:CalcFocus(parent)
	elseif FS.state ~= "castStarted" and FS.state ~= "succeeded" then self:CalcFocus(parent)
	elseif FS.state == "castStarted" and (self.spell == string.sub(GetSpellInfo(FS.spellID["AimS"]), 1, -2) or spell == GetSpellInfo(FS.spellID["AimS"])) then self:CalcFocus(parent)
	elseif self.spellName[0] then self:CalcFocus(parent)
	end
	
	-- if a spell has been queued next cast will be at the end of that spell + delta
	local timeToReg
	if parent ~= "fs" then
		self.now = self.spellEndedAt[1]
		timeToReg = self.spellEndedAt[1] - self.spellSentAt[1]
	elseif self.spellName[0] then
		self.now = self.spellEndedAt[0]
		timeToReg = self.spellEndedAt[0] - self.spellSentAt[0]
	elseif self.spellName[1] then
		self.now = self.spellEndedAt[1]
		timeToReg = self.spellEndedAt[1] - now
	elseif self.spellName[2] and self.spellEndedAt[2] > now then
		self.now = self.spellEndedAt[2]
		timeToReg = self.spellEndedAt[2] - now
	else
		self.now = now
		timeToReg = 0
	end

	-- account for lost or gained ISS buffs
	if FS.talents[FS.spellID["ISS"]] > 0 then
		local steady = GetSpellInfo(FS.spellID["SS"])
		if self.spellName[0] == steady and self.spellName[1] == steady then self.buffs["ISS"] = self.spellEndedAt[0] + 8 + FS.avrgLatency
		elseif self.spellName[1] == steady and self.spellName[2] == steady then self.buffs["ISS"] = self.spellEndedAt[1] + 8 + FS.avrgLatency
		elseif self.spellName[2] == steady and self.spellName[3] == steady and self.spellEndedAt[2] > (now - 2) then self.buffs["ISS"] = self.spellEndedAt[2] + 8 + FS.avrgLatency
		end
	end
	if FS.resources.buffs["ISS"] >= GetTime() and self.buffs["ISS"] < self.now then
		self.haste = self.haste / (1 + FS.talents[FS.spellID["ISS"]] * 0.05)
	elseif FS.resources.buffs["ISS"] < GetTime() and self.buffs["ISS"] >= self.now then
		self.haste = self.haste * (1 + FS.talents[FS.spellID["ISS"]] * 0.05)
	end
	
	self:CalcReg(timeToReg)
end

-- calculate cooldowns and debuff durations when casting given spell
local function CalcTimers(self, parent)
	local spell = self.spell

	-- Cobra Shot
	if spell == GetSpellInfo(FS.spellID["CoS"]) then
		if self.debuffs["SpS"][FS.tarGUID] > self.now then self.debuffs["SpS"][FS.tarGUID] = self.debuffs["SpS"][FS.tarGUID] + 6 end
	-- Explosive Shot
	elseif spell == GetSpellInfo(FS.spellID["ES"]) then
		if self.buffs["LnL"] > self.now then
			local arcs = GetSpellInfo(FS.spellID["AS"])
			local numESAS = self:NumCasts((self.buffs["LnL"] - 12), spell, arcs)
			local numES = self:NumCasts((self.buffs["LnL"] - 12), spell)
			if numESAS < 2 then 
				self.cooldowns["ES"] = 0
				self.lnl = true
			elseif numESAS == 2 then 
				self.cooldowns["ES"] = 0
				self.lnl = true
				self.buffs["LnL"] = 0
				-- prevent clipping only if it's the third ES in a row
				if numES == 2 then self.es = self.now + 2 - FS.avrgLatency - 0.1 end
			else
				self.cooldowns["ES"] = self.now + 6
				self.buffs["LnL"] = 0
			end
			if not self.lnl and self.buffs["BrA"] then
				self.bra = true
				self.buffs["BrA"] = 0
			end
		else
			self.cooldowns["ES"] = self.now + 6
		end
	-- Kill Command
	elseif spell == GetSpellInfo(FS.spellID["KC"]) then
		self.cooldowns["KC"] = self.now + 6
		if self.buffs["KiS"] > self.now then
			self.kis = true
			self.buffs["KiS"] = 0
		end
		if self.buffs["BrA"] then
			self.bra = true
			self.buffs["BrA"] = 0
		end
	-- Chimera Shot
	elseif spell == GetSpellInfo(FS.spellID["CS"]) then
		if self.debuffs["SpS"][FS.tarGUID] > self.now then self.debuffs["SpS"][FS.tarGUID] = self.now + 15 end
		self.cooldowns["CS"] = self.now + 10
		if FS.glyphs[FS.glyphID["CS"]] then self.cooldowns["CS"] = self.cooldowns["CS"] - 1 end
		if self.buffs["BrA"] then
			self.bra = true
			self.buffs["BrA"] = 0
		end
	-- Black Arrow
	elseif spell == GetSpellInfo(FS.spellID["BA"]) then
		self.cooldowns["BA"] = self.now + 30 - FS.talents[FS.spellID["RES"]] * 2
		self.cooldowns["ET"] = self.cooldowns["BA"]
	-- Explosive Trap
	elseif spell == GetSpellInfo(FS.spellID["ET"]) then
		self.cooldowns["ET"] = self.now + 30 - FS.talents[FS.spellID["RES"]] * 2
		self.cooldowns["BA"] = self.cooldowns["ET"]
	-- Aimed Shot
	elseif spell == GetSpellInfo(FS.spellID["AimS"]) then
		if self.buffs["MMM"] > self.now then
			self.mmm = true
			self.buffs["MMM"] = 0
		elseif self.buffs["BrA"] then
			self.bra = true
			self.buffs["BrA"] = 0
		end
	-- Kill Shot
	elseif spell == GetSpellInfo(FS.spellID["KS"]) then
		self.cooldowns["KS"] = self.now + 10
		if FS.glyphs[FS.glyphID["KS"]] then
			local first, last
			for i = 1, 4 do
				if self.spellName[i] == spell then
					if not last then
						last = self.spellSentAt[i]
					else
						first = self.spellSentAt[i]
						break
					end
				end
			end
			if first and (last - first) >= 6 then self.cooldowns["KS"] = 0 end
		end
	-- Serpent Sting
	elseif spell == GetSpellInfo(FS.spellID["SpS"]) then
		self.debuffs["SpS"][FS.tarGUID] = self.now + 15
	-- Hunter's Mark
	elseif spell == GetSpellInfo(FS.spellID["HM"]) then
		self.debuffs["HM"][FS.tarGUID] = self.now + 300
	end
end

-- calculate focus when casting given spell
local function CalcFocus(self, parent)
	local spell = self.spell

	-- Explosive Shot
	if spell == GetSpellInfo(FS.spellID["ES"]) then
		if self.lnl then self.lnl = nil
		elseif self.bra then self.bra = nil
		else self.focus = self.focus - 50 + FS.talents[FS.spellID["EFF"]] * 2
		end
	-- Kill Command
	elseif spell == GetSpellInfo(FS.spellID["KC"]) then
		if not self.bra and self.kis then
			self.kis = nil
			self.focus = self.focus + FS.talents[FS.spellID["KiS"]] * 5
		end
		if self.bra then self.bra = nil
		else self.focus = self.focus - 40 + FS.glyphs[FS.glyphID["KC"]] * 3
		end
	-- Chimera Shot
	elseif spell == GetSpellInfo(FS.spellID["CS"]) then
		if self.bra then self.bra = nil
		else self.focus = self.focus - 50 + FS.talents[FS.spellID["EFF"]] * 2
		end
	-- Black Arrow
	elseif spell == GetSpellInfo(FS.spellID["BA"]) then
		if self.bra then self.bra = nil
		else self.focus = self.focus - 35
		end
	-- Explosive Trap
	elseif spell == GetSpellInfo(FS.spellID["ET"]) then
		self.focus = self.focus - 20 + FS.glyphs[FS.glyphID["TL"]] * 10
	-- Arcane Shot
	elseif spell == GetSpellInfo(FS.spellID["AS"]) then
		if self.buffs["LnL"] > self.now then
			local explo = GetSpellInfo(FS.spellID["ES"])
			local num = self:NumCasts((self.buffs["LnL"] - 12), explo, spell)
			if num >= 2 then self.buffs["LnL"] = 0 end
		elseif self.buffs["BrA"] > self.now then
			self.buffs["BrA"] = 0
		else
			self.focus = self.focus - 25 + FS.talents[FS.spellID["EFF"]]
		end
	-- Aimed Shot
	elseif spell == GetSpellInfo(FS.spellID["AimS"]) or spell == string.sub(GetSpellInfo(FS.spellID["AimS"]), 1, -2) then
		if parent ~= "fs" then self.focus = self.focus - 50
		elseif self.mmm then self.mmm = nil
		elseif self.bra then self.bra = nil
		end
	-- Serpent Sting
	elseif spell == GetSpellInfo(FS.spellID["SpS"]) then
		if self.bra then self.bra = nil
		else self.focus = self.focus - 25
		end
	-- Steady and Cobra for button2
	elseif parent ~= "fs" and (spell == GetSpellInfo(FS.spellID["CoS"]) or spell == GetSpellInfo(FS.spellID["SS"])) then
		self.focus = self.focus + self.gains
	end
end

-- returns focus after a given time to regenerate at a given startvalue
local function CalcReg(self, timeToReg)
	self.focus = self.focus + 4 * timeToReg * self.relevantHaste
	if FS.talents[FS.spellID["KinS"]] > 0 and self.focus > (100  + FS.talents[FS.spellID["KinS"]] * 5) then self.focus = 100 + FS.talents[FS.spellID["KinS"]] * 5
	elseif FS.talents[FS.spellID["KinS"]] == 0 and self.focus > 100 then self.focus = 100
	end
end

-- stabilize recommendation by making it harder to switch to a different spell once it has been determined at castSent
local function Damping(self, acrn, minFocus, maxFocus)
	if not self.recom or FS.state == "castSent" then
		return minFocus, maxFocus
	elseif self.recom == GetSpellInfo(FS.spellID[acrn]) then
		if minFocus and minFocus > 0 then minFocus = minFocus - FS.db.profile.damping end
		if maxFocus and maxFocus < 110 then maxFocus = maxFocus + FS.db.profile.damping end
	else
		if minFocus and minFocus > 0 then minFocus = minFocus + FS.db.profile.damping end
		if maxFocus and maxFocus < 110 then maxFocus = maxFocus - FS.db.profile.damping end
	end
	return minFocus, maxFocus
end

-- return the amount of focus that is regenerated (w/o SS/CoS) until given spell is off CD again or debuff run off
local function RegTilRdy(self, acrn)
	local dur
	if self.cooldowns[acrn] and self.cooldowns[acrn] >= self.now then dur = self.cooldowns[acrn] - self.now
	elseif self.debuffs[acrn] and self.debuffs[acrn][FS.tarGUID] and self.debuffs[acrn][FS.tarGUID] >= self.now then dur = self.debuffs[acrn][FS.tarGUID] - self.now
	else return 0
	end
	return (4 * self.relevantHaste * dur)
end

-- return true if player would be overcapped after num SS/CoS or cost free shots
local function Overcapped(self, isReg, num)
	local regen, ctime
	
	-- if either SS or CoS is used add the gains
	if isReg then
		ctime = 2 / self.haste
		if not FS.sets then FS:GetItems() end
		if FS.sets.T11 and FS.sets.T11 >= 4 then ctime = ctime - 0.2 end
		regen = ctime * self.relevantHaste * 4 + self.gains
	else regen = 4 * self.relevantHaste
	end
	
	-- possibly add a termination bonus
	if isReg and self.tarHP <= 0.25 then regen = regen + FS.talents[FS.spellID["TER"]] * 3 end

	-- if num is omitted assume 1
	if not num then num = 1 end
	regen = num * regen
	
	-- if spec is MM and AimS is an option account for the time AimS needs to cast after finishing the SS/CoS
	if isReg and FS.spec == "MM" then
	-- Determine the current useStyle for AimS
		local useStyle
		local useAimed = true
		ctime = 2.9 / self.haste
		if self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then
			if ctime <= FS.db.profile.aims.caBreakpoint then useStyle = FS.db.profile.aims.caUseStyleUnder
			else useStyle = FS.db.profile.aims.caUseStyleOver
			end
		else
			if ctime <= FS.db.profile.aims.noCaBreakpoint then useStyle = FS.db.profile.aims.noCaUseStyleUnder
			else useStyle = FS.db.profile.aims.noCaUseStyleOver
			end
		end
		
		-- Never use AimS -> return false
		if useStyle == "never" then useAimed = false
		-- CS > AimS -> return false if CS CD free or focus needs to be saved in order to cast CS in time
		elseif useStyle == "prioUnderCS" and self.cooldowns["CS"] <= (self.now + 3) and FS.cost["CS"] > (self.focus + self:RegTilRdy("CS") - FS.cost["AimS"]) then useAimed = false
		end
		
		if useAimed then 
			regen = regen + ctime * self.relevantHaste * 4 + self.gains
		end
	end
	
	if (self.focus + regen) > (100 + FS.talents[FS.spellID["KinS"]] * 5) then return true
	else return false
	end
end

-- returns true if the last shot was the given one
local function IsLastShot(self, spell)
	if self.spellName[0] and self.spellName[0] == spell then return true
	elseif self.spellName[1] and self.spellName[1] == spell then return true
	elseif self.spellName[2] and self.spellName[2] == spell then return true
	else return false
	end
end

-- returns true if the next to last shot was the given one
local function IsNextToLastShot(self, spell)
	if self.spellName[0] and self.spellName[1] == spell then return true
	elseif self.spellName[1] and self.spellName[2] == spell then return true
	elseif self.spellName[2] and self.spellName[3] == spell then return true
	else return false
	end
end

-- return true if next shot should be a SS to get an ISS proc (with or without overcap prevention)
local function CheckISS(self)
	local steady = GetSpellInfo(FS.spellID["SS"])
	if self:IsLastShot(steady) and not self:IsNextToLastShot(steady) then return true
	elseif self.buffs["ISS"] >= self.now and self.buffs["ISS"] <= (self.now + 4) then return true
	else return false
	end
end

-- returns true if KC is off CD, RiF is up and focus meets given requirements
local function CheckRiF(self, ...)
	local minFocus, maxFocus = ...
	if self.buffs["RiF!"] < self.now or not UnitExists("pet") or UnitIsDead("pet") then return false end
	if not minFocus then minFocus = FS.cost["KC"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("KC", minFocus, maxFocus)
	return self.cooldowns["KC"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if either MfD or HM is on target and doesnt run out shortly or check has been turned off
local function IsHMonTarget(self)
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS.db.profile.hm.suggestHM == "never" then return true
	elseif FS.db.profile.hm.suggestHM == "ooc" and not FS.ooc then return true
	elseif FS.db.profile.hm.suggestHM == "level" and UnitLevel("target") ~= -1 then return true
	elseif FS.db.profile.hm.suggestHM == "oocANDlevel" and (not FS.ooc or UnitLevel("target") ~= -1) then return true
	elseif FS.db.profile.hm.suggestHM == "oocORlevel" and not FS.ooc and UnitLevel("target") ~= -1 then return true
	elseif FS_TTL and FS_TTL.delta < FS.db.profile.hm.minTTLHM then return true
	end
	return self.debuffs["MfD"][FS.tarGUID] >= self.now or self.debuffs["HM"][FS.tarGUID] >= (self.now + FS.db.profile.hm.minTimeHM)
end

-- returns true if ES is off CD, DoT isnt ticking on target and focus meets given requirements
local function CheckES(self, ...)
	local minFocus, maxFocus = ...
	if self.buffs["LnL"] > self.now then minFocus = 0
	else minFocus = FS.cost["ES"]
	end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("ES", minFocus, maxFocus)
	return self.cooldowns["ES"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.es <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if CS is off CD and focus meets given requirements
local function CheckCS(self, ...)
	local useStyle
	local ctime = 2.9 / self.haste
	-- Determine the useStyle
	if self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then
		if ctime <= FS.db.profile.aims.caBreakpoint then useStyle = FS.db.profile.aims.caUseStyleUnder
		else useStyle = FS.db.profile.aims.caUseStyleOver
		end
	else
		if ctime <= FS.db.profile.aims.noCaBreakpoint then useStyle = FS.db.profile.aims.noCaUseStyleUnder
		else useStyle = FS.db.profile.aims.noCaUseStyleOver
		end
	end

	-- Ignore CS -> return false
	if useStyle == "ignoreCS" then return false
	-- AimS > CS but keep the sting up -> false if AimS can be cast without SpS running risk of not beeing refreshed in time
	elseif useStyle == "prioOverCS" then
		-- if there's no SpS on target there's no need to refresh it either -> don't suggest CS
		if self.debuffs["SpS"][FS.tarGUID] <= self.now then	return false end
		-- if there IS a SpS on target calculate the leftover focus considering AimS cost and regenerated focus while SpS is ticking	
		local leftover = self.focus + self:RegTilRdy("SpS")
		if self.buffs["MMM"] <= self.now then leftover = leftover - FS.cost["AimS"] end
		-- if leftover focus is sufficiant to cast CS in time and SpS doesn't run out in less than 5s don't -> don't suggest CS
		if leftover > FS.cost["CS"] and (self.now + 5) < self.debuffs["SpS"][FS.tarGUID] then
			return false
		end
	end
	
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = FS.cost["CS"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("CS", minFocus, maxFocus)
	return self.cooldowns["CS"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if there's a MMM proc or focus meets requirement
local function CheckAimS(self, ...)
	local minFocus, maxFocus = ...
	if self.buffs["MMM"] > self.now then
		self.buffs["MMM"] = 0
		self.mmm = true
		return true
	end
	
	-- force using AS instead of AimS as focus dump if 4pT13 bonus can proc
	if FS.db.profile.T13.forceAS and FS.sets.T13 and FS.sets.T13 >= 4 and (self.now - FS.lastT13procAt) > 110 then
		return false
	end
	
	local useStyle
	local ctime = 2.9 / self.haste
	-- Determine the current useStyle
	if self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then
		if ctime <= FS.db.profile.aims.caBreakpoint then useStyle = FS.db.profile.aims.caUseStyleUnder
		else useStyle = FS.db.profile.aims.caUseStyleOver
		end
	else
		if ctime <= FS.db.profile.aims.noCaBreakpoint then useStyle = FS.db.profile.aims.noCaUseStyleUnder
		else useStyle = FS.db.profile.aims.noCaUseStyleOver
		end
	end
	
	-- Never use AimS -> return false
	if useStyle == "never" then return false
	-- CS > AimS -> return false if CS CD free or focus needs to be saved in order to cast CS in time
	elseif useStyle == "prioUnderCS" and self.cooldowns["CS"] <= (self.now + 3) and FS.cost["CS"] > (self.focus + self:RegTilRdy("CS") - FS.cost["AimS"]) then return false
	end
	
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = FS.cost["AimS"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("AimS", minFocus, maxFocus)
	return self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if BA is off CD and focus meets given requirements
local function CheckBA(self, ...)
	if FS.db.profile.ba.suggestBA == "never" or (FS.db.profile.ba.suggestBA == "level" and UnitLevel("target") ~= -1) then return false end
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.ba.minTTLBA then return false end
	local minFocus, maxFocus = ...
	if not minFocus or not FS.db.profile.ba.saveFocusBA then minFocus = FS.cost["BA"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("BA", minFocus, maxFocus)
	return self.cooldowns["BA"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if focus meets given requirements
local function CheckET(self, ...)
	if FS.db.profile.ba.suggestET == "never" or (FS.db.profile.ba.suggestET == "level" and UnitLevel("target") ~= -1) then return false end
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.ba.minTTLET then return false end
	local minFocus, maxFocus = ...
	if not minFocus or not FS.db.profile.ba.saveFocusET then minFocus = FS.cost["TL"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("ET", minFocus, maxFocus)
	return self.cooldowns["ET"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if KS is off CD and focus meets given requirements
local function CheckKS(self, ...)
	local maxFocus = ...
	if not maxFocus then maxFocus = 110 end
	_, maxFocus = self:Damping("KS", nil, maxFocus)
	return self.cooldowns["KS"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.tarHP <= 0.2 and self.focus <= maxFocus
end

-- returns true if SpS isnt on target already and self.focus meets the given requirements
local function CheckSpS(self, ...)
	if FS.db.profile.sps.suggestSpS == "never" or (FS.db.profile.sps.suggestSpS == "level" and UnitLevel("target") ~= -1) then return false end
	if FS.spec == "MM" then
		local ctime = 2.9 / self.haste
		-- if useStyle suggests to ignore CS dont cast SpS either
		if self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then
			if ctime <= FS.db.profile.aims.caBreakpoint and FS.db.profile.aims.caUseStyleUnder == "ignoreCS" then return false
			elseif ctime > FS.db.profile.aims.caBreakpoint and FS.db.profile.aims.caUseStyleOver == "ignoreCS" then return false
			end
		else
			if ctime <= FS.db.profile.aims.noCaBreakpoint and FS.db.profile.aims.noCaUseStyleUnder == "ignoreCS" then return false
			elseif ctime > FS.db.profile.aims.noCaBreakpoint and FS.db.profile.aims.noCaUseStyleOver == "ignoreCS" then return false
			end
		end
	end
	
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.sps.minTTLSpS then return false end
	local minFocus, maxFocus = ...
	if not minFocus or not FS.db.profile.sps.saveFocusSpS then minFocus = FS.cost["SpS"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("SpS", minFocus, maxFocus)
	return self.debuffs["SpS"][FS.tarGUID] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if focus requirements are met and theres no LnL proc that should be saved for ES
local function CheckAS(self, ...)
	if FS.spec == "MM" then

		-- force using AS overrules CA
		if not (FS.db.profile.T13.forceAS and FS.sets.T13 and FS.sets.T13 >= 4 and (self.now - FS.lastT13procAt) > 110) then
			-- check wether AimS or AS is used as focus dump
			local ctime = 2.9 / self.haste
			if self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then
				if ctime <= FS.db.profile.aims.caBreakpoint and FS.db.profile.aims.caUseStyleUnder ~= "never" then return false
				elseif ctime > FS.db.profile.aims.caBreakpoint and FS.db.profile.aims.caUseStyleOver ~= "never" then return false
				end
			elseif (self.tarHP < 0.9 or FS.talents[FS.spellID["CA"]] == 0) then 
				if ctime <= FS.db.profile.aims.noCaBreakpoint and FS.db.profile.aims.noCaUseStyleUnder ~= "never" then return false
				elseif ctime > FS.db.profile.aims.noCaBreakpoint and FS.db.profile.aims.noCaUseStyleOver ~= "never" then return false
				end
			end
		end
		-- if T12 proc is up check if casting SS will overcap us before CS is ready
		-- T12 proc wird in CalcFocus gelscht!?!?!
	end
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = FS.cost["AS"] end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("AS", minFocus, maxFocus)
	return self.buffs["LnL"] < self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if KC is off CD and focus meets given requirements
local function CheckKC(self, ...)
	local minFocus, maxFocus = ...
	if not UnitExists("pet") or UnitIsDead("pet") then return false end
	if self.buffs["KiS"] > self.now then minFocus = FS.cost["KC"] - FS.talents[FS.spellID["KiS"]] * 5
	else minFocus = FS.cost["KC"]
	end
	if FS.spec == "SV" and not FS.db.profile.lnl.useKCasdump then return false end
	if not maxFocus then maxFocus = 110 end
	minFocus, maxFocus = self:Damping("KC", minFocus, maxFocus)
	return self.cooldowns["KC"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if casting doesnt lead to overcap or player has got CA and target is above 90% in which case CoS is still better than AS
local function CheckCoS(self, ...)
	if FS.level <= 80 then return false -- player hasnt got CoS yet
	--elseif self.tarHP >= 0.9 and FS.talents[FS.spellID["CA"]] > 0 then return true -- CoS is worthwhile to cast with +60% crit no matter what
	else return not self:Overcapped(true)
	end
end

-- pick the most beneficial spell based on given resources
local function PickSpell(self, btnName)
	local focus, spell = self.focus

	FS:GetSpellCosts()
	if self.buffs["BrA"] > self.now then for key, val in pairs(FS.cost) do FS.cost[key] = 0 end
	elseif self.buffs["TBW"] > self.now then for key, val in pairs(FS.cost) do FS.cost[key] = val * 0.5 end
	end
	
	self.recom = FS[btnName].recom

	-- always cast HM on target if ooc regardles of spec
	if FS.ooc and not self:IsHMonTarget() then
		spell = FS.spellID["HM"]
	-- SV rotation
	elseif FS.spec == "SV" then
		-- Kill Shot
		if self:CheckKS() and not self:Overcapped() then
			spell = FS.spellID["KS"]
		-- Explosive Shot
		elseif self:CheckES() then
			spell = FS.spellID["ES"]
		-- Explosive Trap
		elseif FS.db.profile.ba.useTraps ~= "never" and self:CheckET((FS.cost["ES"] + FS.cost["TL"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["ET"]
		-- Black Arrow
		elseif self:CheckBA((FS.cost["ES"] + FS.cost["BA"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["BA"]
		-- Serpent Sting
		elseif self:CheckSpS((FS.cost["ES"] + FS.cost["SpS"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["SpS"]
		-- Hunter's Mark
		elseif not self:IsHMonTarget() and not self:Overcapped() then
			spell = FS.spellID["HM"]
		-- Arcane Shot
		elseif self:CheckAS((FS.cost["ES"] + FS.cost["BA"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["AS"]
		-- Cobra Shot
		elseif self:CheckCoS() then	
			spell = FS.spellID["CoS"]
		-- Kill Command
		elseif self:CheckKC() then
			spell = FS.spellID["KC"]
		-- Steady Shot
		elseif FS.level <= 80 then
			spell = FS.spellID["SS"]
		-- Cobra Shot
		else
			spell = FS.spellID["CoS"]
		end
	-- BM rotation
	elseif FS.spec == "BM" then
		-- Kill Shot
		if self:CheckKS() and not self:Overcapped() then
			spell = FS.spellID["KS"]
		-- Kill Command
		elseif self:CheckKC() then
			spell = FS.spellID["KC"]
		-- Serpent Sting
		elseif self:CheckSpS((FS.cost["KC"] + FS.cost["SpS"] - self:RegTilRdy("KC"))) then
			spell = FS.spellID["SpS"]
		-- Explosive Trap
		elseif FS.db.profile.ba.useTraps == "always" and self:CheckET((FS.cost["KC"] + FS.cost["TL"] - self:RegTilRdy("KC"))) then
			spell = FS.spellID["ET"]
		-- Hunter's Mark
		elseif not self:IsHMonTarget() and not self:Overcapped() then
			spell = FS.spellID["HM"]
		-- Arcane Shot
		elseif self:CheckAS((FS.cost["KC"] + FS.cost["SpS"] - self:RegTilRdy("KC"))) then
			spell = FS.spellID["AS"]
		-- Cobra Shot
		elseif FS.level > 80 then	
			spell = FS.spellID["CoS"]
		-- Steady Shot
		else
			spell = FS.spellID["SS"]
		end
	-- MM rotation
	elseif FS.spec == "MM" then
		-- Steady Shot
		if self:CheckISS(true) then
			spell = FS.spellID["SS"]
		-- Kill Shot
		elseif self:CheckKS() and not self:Overcapped() then
			spell = FS.spellID["KS"]
		-- Chimera Shot
		elseif self:CheckCS() then
			spell = FS.spellID["CS"]
		-- Serpent Sting
		elseif self:CheckSpS((FS.cost["CS"] + FS.cost["SpS"] - self:RegTilRdy("CS"))) then
			spell = FS.spellID["SpS"]
		-- Aimed Shot
		elseif self:CheckAimS() then
			spell = FS.spellID["AimS"]
		-- Kill Command
		elseif self:CheckRiF() and not self:Overcapped() then
			spell = FS.spellID["KC"]
		-- Hunter's Mark
		elseif not self:IsHMonTarget() and not self:Overcapped() then
			spell = FS.spellID["HM"]
		-- Arcane Shot
		elseif self:CheckAS((FS.cost["CS"] + FS.cost["SpS"] - self:RegTilRdy("CS"))) then
			spell = FS.spellID["AS"]
		-- Steady Shot
		else
			spell = FS.spellID["SS"]
		end
	-- no spec
	else
		-- Kill Shot
		if self:CheckKS() then
			spell = FS.spellID["KS"]
		-- Serpent Sting
		elseif self:CheckSpS() then
			spell = FS.spellID["SpS"]
		-- Explosive Trap
		elseif FS.db.profile.ba.useTraps == "allTraps" and self:CheckET((FS.cost["SpS"] + FS.cost["TL"])) then
			spell = FS.spellID["ET"]
		-- Hunter's Mark
		elseif not self:IsHMonTarget() and not self:Overcapped() then
			spell = FS.spellID["HM"]
		-- Arcane Shot
		elseif self:CheckAS((FS.cost["SpS"] + FS.cost["AS"])) then
			spell = FS.spellID["AS"]
		-- Cobra Shot
		elseif FS.level > 80 then	
			spell = FS.spellID["CoS"]
		-- Steady Shot
		else
			spell = FS.spellID["SS"]
		end
	end

	local spellName = GetSpellInfo(spell)
	
	-- draw the corrosponding texture
	if self.spellName[0] then
		table.insert(self.spellName, 1, self.spellName[0])
		table.remove(self.spellName, 5)
		table.insert(self.spellSentAt, 1, self.spellSentAt[0])
		table.remove(self.spellSentAt, 5)
		table.insert(self.spellStartedAt, 1, self.spellStartedAt[0])
		table.remove(self.spellStartedAt, 5)
		table.insert(self.spellEndedAt, 1, self.spellEndedAt[0])
		table.remove(self.spellEndedAt, 5)
		self.spellName[0] = nil
		self.spellSentAt[0] = nil
		self.spellStartedAt[0] = nil
		self.spellEndedAt[0] = nil
	end
	if self.spellName[1] then
		table.insert(self.spellName, 1, spellName)
		table.remove(self.spellName, 5)
		table.insert(self.spellSentAt, 1, self.now)
		table.remove(self.spellSentAt, 5)
		table.insert(self.spellStartedAt, 1, self.now + FS.avrgLatency)
		table.remove(self.spellStartedAt, 5)
		table.insert(self.spellEndedAt, 1, self.now + FS:GetCastDelta(spellName, self.haste, self.buffs["MMM"], self.now)) --> new self.now for next button - ToDo: get it as precise as possible
		table.remove(self.spellEndedAt, 5)
	else
		self.spellName[1] = spellName
		self.spellSentAt[1] = self.now
		self.spellStartedAt[1] = self.now + FS.avrgLatency
		self.spellEndedAt[1] = self.now + FS:GetCastDelta(spellName, self.haste, self.buffs["MMM"], self.now) --> new self.now for next button - ToDo: get it as precise as possible
	end
	FS[btnName].btn.texture:SetTexture(GetSpellTexture(spell))
	FS[btnName]:DrawCDF(FS:GetGCD())
	FS[btnName].recom = GetSpellInfo(spell)
	local inRange
	if spell == FS.spellID["AimS"] or spell == FS.spellID["ET"] then inRange = IsSpellInRange(GetSpellInfo(FS.spellID["AutoS"]), "target")
	else inRange = IsSpellInRange(GetSpellInfo(spell), "target")
	end
	if inRange == 1 then FS[btnName].btn.texture:SetVertexColor(1, 1, 1)
	else FS[btnName].btn.texture:SetVertexColor(1, 0.15, 0.15)
	end
end

-- creates an resource object 
function FS:CreateResources()
	local obj = {
		now = 0,
		focus = 100,
		es = 0,
		haste = 1,
		tarHP = 0,
		spell = nil,
		spellName = {},
		spellSentAt = {},
		spellStartedAt = {},
		spellEndedAt = {},
		charges = {
			["LnL"] = 0,
		},
		buffs = {
			["LnL"] = 0,
			["MMM"] = 0,
			["KiS"] = 0,
			["ISS"] = 0,
			["TBW"] = 0,
			["RiF!"] = 0,
			["HP"] = 0,
			["IIT"] = 0,
			["WFT"] = 0,
			["BrA"] = 0,
			["CHR"] = 0,
		},
		debuffs = {
			["HM"] = {},
			["SpS"] = {},
			["MfD"] = {},
		},
		cooldowns = {
			["CS"] = 0,
			["BA"] = 0,
			["ES"] = 0,
			["KS"] = 0,
			["AimS"] = 0,
			["KC"] = 0,
			["ET"] = 0,
		},
		focusGains = {
			["CoS"] = 0,
			["SS"] = 0,
			["RiF"] = 0,
			["AimS"] = 0,
		},
		lastCast = {
			["HM"] = 0,
			["SpS"] = 0,
			["CS"] = 0,
			["BA"] = 0,
			["ES"] = 0,
			["KS"] = 0,
			["AimS"] = 0,
			["KC"] = 0,
			["ET"] = 0,
			["CoS"] = 0,
			["SS"] = 0,
		},
		NumCasts = NumCasts,
		CheckES = CheckES,
		CheckCS = CheckCS,
		CheckBA = CheckBA,
		CheckET = CheckET,
		CheckKS = CheckKS,
		CheckSpS = CheckSpS,
		CheckAimS = CheckAimS,
		CheckAS = CheckAS,
		CheckKC = CheckKC,
		CheckCoS = CheckCoS,
		CheckISS = CheckISS,
		CheckRiF = CheckRiF,
		IsLastShot = IsLastShot,
		IsNextToLastShot = IsNextToLastShot,
		IsHMonTarget = IsHMonTarget,
		Damping = Damping,
		RegTilRdy = RegTilRdy,
		Overcapped = Overcapped,
		PickSpell = PickSpell,
		GetResources = GetResources,
		SetResources = SetResources,
		FocusGains = FocusGains,
		GetBuffs = GetBuffs,
		GetDebuffs = GetDebuffs,
		PurgeDebuffs = PurgeDebuffs,
		CastSpell = CastSpell,
		CalcTimers = CalcTimers,
		CalcFocus = CalcFocus,
		CalcReg = CalcReg,
	}
	return obj
end

------------
--- button ---
------------

-- draw the cooldownframe or animate the frame in some style
local function DrawCDF(self, start, duration)
	local now = GetTime()
	--local FS_ANI = FS:GetModule("Animation", true)
	--if FS_ANI:IsEnabled() and FS.db.profile.animation.style ~= "none" then return false end
	if FS.isInGCD ~= (start + duration) and (start + duration) > 0 then
		self.cdf:SetCooldown(start, duration)
		self.cdf:Show()
		FS.isInGCD = start + duration
	elseif FS.isInGCD and FS.isInGCD > now and (start + duration) == 0 then
		self.cdf:Hide()
		FS.isInGCD = false
	elseif FS.isInGCD and (start + duration) == 0 then
		FS.isInGCD = false
	end
end

-- hides the objects frame
local function Hide(self)
	self.btn:Hide()
	self.cdf:Hide()
end

-- shows the objects frame
local function Show(self)
	self.btn:Show()
	if FS.isInGCD then self.cdf:Show() end
end

-- creates an button object
function FS:CreateButton(name)
	local obj = {
		name = name,
		resources = self:CreateResources(),
		mov = CreateFrame("Frame", name.."Mover", UIParent),
		btn = CreateFrame("Button", name.."Button", mov),
		cdf = CreateFrame("Cooldown", name.."CooldownFrame", btn),
		PickSpell = PickSpell,
		DrawCDF = DrawCDF,
		Hide = Hide,
		Show = Show,
	}
	
	obj.mov:SetFrameStrata("BACKGROUND")
	obj.mov:SetWidth(40)
	obj.mov:SetHeight(40)
	obj.mov:SetScale(self.db.profile[name]["scale"]/100)
	obj.mov:SetPoint(self.db.profile[name]["point"], self.db.profile[name]["x"], self.db.profile[name]["y"])
	obj.mov:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 32,})
	obj.mov:SetBackdropColor(0, 0, 0, 0.4)
	obj.mov:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	obj.mov:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() FS.db.profile[name]["point"], _, _, FS.db.profile[name]["x"], FS.db.profile[name]["y"] = self:GetPoint() end)
	obj.mov:SetMovable(true)
	obj.mov:SetClampedToScreen(true)
	obj.mov.str = obj.mov:CreateFontString("String", "ARTWORK", "GameFontNormal")
	obj.mov.str:SetFont(GameFontNormal:GetFont(), 10)
	obj.mov.str:SetJustifyH("CENTER")
	obj.mov.str:SetJustifyV("TOP")
	obj.mov.str:SetPoint("CENTER", obj.mov, "CENTER", 0, 30)
	obj.mov.str:SetTextColor(1, 1, 1, 1)
	obj.mov.str:SetText(name)

	-- hide anchor frame if it is locked
	if self.db.profile.lock then
		obj.mov:SetAlpha(0)
		obj.mov:EnableMouse(false)
	-- otherwise show and make it moveable
	else
		obj.mov:SetAlpha(1)
		obj.mov:EnableMouse(true)
	end
	obj.btn:SetWidth(32)
	obj.btn:SetHeight(32)
	obj.btn:EnableMouse(false)
	obj.btn:SetAlpha(self.db.profile[name]["alpha"]/100)
	obj.btn:SetScale(self.db.profile[name]["scale"]/100)
	obj.btn:SetPoint("TOPLEFT", obj.mov)
	obj.btn.texture = obj.btn:CreateTexture(nil, "BACKGROUND")
	obj.btn.texture:SetTexture(nil)
	obj.btn.texture:ClearAllPoints()
	obj.btn.texture:SetAllPoints(obj.btn)
	obj.cdf:SetHeight(32)
	obj.cdf:SetWidth(32)
	obj.cdf:ClearAllPoints()
	obj.cdf:SetAllPoints(obj.btn)
	obj.cdf:SetAlpha(1)

	return obj
end

-------------
--- Debug ---
-------------

-- makes sure that the given value is concatenable
function FS:mkStr(val, round, modula)
	local out = val
	if val == nil then out = "nil"
	elseif val == false then out = "false"
	elseif type(val) ~= "number" and type(val) ~= "string" then out = type(val)
	end
	if modula and tonumber(out) then out = val % modula end
	if round and tonumber(out) then out = FS:Round(out, round) end
	return out
end

-- takes a snapshot of current resources and logs them
function FS:GetLogEntry(name)
	local entry = {
		name = name,
		recom = FS.button1.recom,
		state = FS.state,
		times = "real now: "..GetTime().." / self.now: "..FS:Round(FS.button1.resources.now, 3).." / lastCastSentAt: "..self:mkStr(FS.lastCastSentAt).." / lastCastStartedAt: "..self:mkStr(FS.lastCastStartedAt).." / lastCastSucceededAt: "..self:mkStr(FS.lastCastSucceededAt),
		focus = "real focus: "..UnitPower("player").." / self.focus: "..FS:Round(FS.button1.resources.focus, 3),
		spellNames = "sNam[0]: "..self:mkStr(FS.spellName[0]).." / sNam[1]: "..self:mkStr(FS.spellName[1]).." / sNam[2]: "..self:mkStr(FS.spellName[2]).." / sNam[3]: "..self:mkStr(FS.spellName[3]).." / sNam[4]: "..self:mkStr(FS.spellName[4]),
		spellSentAt = "sSend[0]: "..self:mkStr(FS.spellSentAt[0], 3, 100).." / sSend[1]: "..self:mkStr(FS.spellSentAt[1], 3, 100).." / sSend[2]: "..self:mkStr(FS.spellSentAt[2], 3, 100).." / sSend[3]: "..self:mkStr(FS.spellSentAt[3], 3, 100).." / sSend[4]: "..self:mkStr(FS.spellSentAt[4], 3, 100),
		spellStartedAt = "sStart[0]: "..self:mkStr(FS.spellStartedAt[0], 3, 100).." / sStart[1]: "..self:mkStr(FS.spellStartedAt[1], 3, 100).." / sStart[2]: "..self:mkStr(FS.spellStartedAt[2], 3, 100).." / sStart[3]: "..self:mkStr(FS.spellStartedAt[3], 3, 100).." / sStart[4]: "..self:mkStr(FS.spellStartedAt[4], 3, 100),
		spellEndedAt = "sEnd[0]: "..self:mkStr(FS.spellEndedAt[0], 3, 100).." / sEnd[1]: "..self:mkStr(FS.spellEndedAt[1], 3, 100).." / sEnd[2]: "..self:mkStr(FS.spellEndedAt[2], 3, 100).." / sEnd[3]: "..self:mkStr(FS.spellEndedAt[3], 3, 100).." / sEnd[4]: "..self:mkStr(FS.spellEndedAt[4], 3, 100),
		misc = "gcd: "..FS.gcd.." / inGCD: "..self:mkStr(FS.inGCD).." / es: "..self:mkStr(FS.button1.resources.es).." / haste = "..self:mkStr(FS.button1.resources.haste),
		regTilRdy = "CS: "..FS.button1.resources:RegTilRdy("CS").." / KC: "..FS.button1.resources:RegTilRdy("KC").." / ES: "..FS.button1.resources:RegTilRdy("ES"),
		charLnL = FS.button1.resources.charges["LnL"],
		debuffs = "SpS: "..self:mkStr(FS.button1.resources.debuffs["SpS"][FS.tarGUID]).." / HM: "..self:mkStr(FS.button1.resources.debuffs["HM"][FS.tarGUID]).." / MfD: "..self:mkStr(FS.button1.resources.debuffs["MfD"][FS.tarGUID]),
		buffs = "LnL: "..FS.button1.resources.buffs["LnL"].." / MMM: "..FS.button1.resources.buffs["MMM"].." / KiS: "..FS.button1.resources.buffs["KiS"].." / ISS: "..FS.button1.resources.buffs["ISS"].." / TBW: "..FS.button1.resources.buffs["TBW"].." / RiF!: "..FS.button1.resources.buffs["RiF!"],
		cooldowns = "CS: "..FS.button1.resources.cooldowns["CS"].." / KC: "..FS.button1.resources.cooldowns["KC"].." / ES: "..FS.button1.resources.cooldowns["ES"].." / ET: "..FS.button1.resources.cooldowns["ET"].." / BA: "..FS.button1.resources.cooldowns["BA"],
		focusGains = "SS: "..FS.resources.focusGains["SS"].." / CoS: "..FS.resources.focusGains["CoS"].." / RiF: "..FS.resources.focusGains["RiF"],
		issCheck = FS.issCheck,
		debuffCheck = FS.debuffCheck,
	}
	if not FS.db.profile.log then FS.db.profile.log = {} end
	table.insert(FS.db.profile.log, entry)
end

-- writes the snapshot to the savedvariables
function FS:WriteLogEntry(name)
	if type(self.log) ~= "table" then self.log = { lastSpell, lastState } end
	if FS.button1.recom ~= self.log.lastSpell or FS.state ~= self.log.lastState then
		self:GetLogEntry(name)
		if name == "After" then
			self.log.lastSpell = FS.button1.recom
			self.log.lastState = FS.state
		end
	end
end

-- creates the string for the debug frame
local function GetDebugStr(self)
	local str = self.str
	local i, a
	
	-- FS
	str.fs = "State: "..FS.state.."\n\n"
	for i = 0, 4 do
		str.fs = str.fs.."spell["..i.."]: "..FS:mkStr(FS.spellName[i]).." / "..FS:mkStr(FS.spellSentAt[i], 3, 100).." / "..FS:mkStr(FS.spellStartedAt[i], 3, 100).." / "..FS:mkStr(FS.spellEndedAt[i], 3, 100).."\n"
	end
	str.fs = str.fs.."\nnow: "..FS:Round((FS.resources.now % 60), 2).."\n"
	if FS.resources.debuffs["SpS"][FS.tarGUID] and FS.resources.now then a = FS:Round((FS.resources.debuffs["SpS"][FS.tarGUID] - FS.resources.now), 2) end
	if not a then a = "" elseif a < 0 then a = 0 end
	str.fs = str.fs.."SpS: "..a.."\n"
	str.fs = str.fs.."focus: "..UnitPower("player").."\n"
	if FS.isInGCD then str.fs = str.fs.."isInGCD: "..FS.isInGCD.."\n" else str.fs = str.fs.."isInGCD: false\n" end
	str.fs = str.fs.."latency: "..FS:Round(FS.avrgLatency, 3).."\n\n"
	str.fs = str.fs.."focusGains[SS]: "..FS.resources.focusGains["SS"].."\n"
	str.fs = str.fs.."focusGains[CoS]: "..FS.resources.focusGains["CoS"].."\n"
	str.fs = str.fs.."focusGains[RiF]: "..FS.resources.focusGains["RiF"].."\n"	
	str.fs = str.fs.."focusGains[AimS]: "..FS.resources.focusGains["AimS"].."\n"	
	-- button 1-2
	for i = 1, 2 do
		local n = "button"..i
		if FS.db.profile[n].enabled then
			local a
			str[n] = "Frame: "..n.."\n\n"
			if FS[n].resources.spell then str[n] = str[n].."recom: "..FS[n].recom.."\n"
			else str[n] = str[n].."recom:\n"
			end
			if FS[n].resources.now then str[n] = str[n].."now: "..FS:Round((FS[n].resources.now % 60), 2).."\n"
			else str[n] = str[n].."now:\n"
			end
			if FS[n].resources.debuffs["SpS"][FS.tarGUID] and FS[n].resources.now then a = FS:Round((FS[n].resources.debuffs["SpS"][FS.tarGUID] - FS[n].resources.now), 2) end
			if not a then a = "" elseif a < 0 then a = 0 end
			str[n] = str[n].."SpS: "..a.."\n"
			if FS[n].resources.focus then str[n] = str[n].."focus: "..FS:Round(FS[n].resources.focus, 2).."\n"
			else str[n] = str[n].."focus:\n"
			end
		end
	end
end

-- writes the text at the frame
local function WriteText(self)
	local i
	self.btn.fs:SetPoint("TOPLEFT", self.btn, "TOPLEFT", 20, -15)
	self.btn.fs:SetText(self.str.fs)
	local width, height, num, x, y = 250, 220
	for i = 1, 2 do
		local n = "button"..i
		if FS.db.profile[n].enabled then
			num = i
			if i < 3 then x = i * width + 20 else x = (i - 2) * width + 20 end
			if i < 3 then y = -15 else y = -15 - height end
			self.btn[n]:SetPoint("TOPLEFT", self.btn, "TOPLEFT", x, y)
			self.btn[n]:SetText(self.str[n])
		end
	end
	if num == 1 then x = 2 * width + 20 else x = 3 * width + 20 end
	y = height - 15
	self.btn:SetWidth(x)
	self.btn:SetHeight(y)
end

-- create an frame object for debugging information
function FS:CreateDebugFrame()
	local obj = {
		btn = CreateFrame("Button", btn ,UIParent),
		str = {},
		GetDebugStr = GetDebugStr,
		WriteText = WriteText,
	}
	obj.btn:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", tile = true, tileSize = 32, edgeSize = 32, insets = {left = 11, right = 12, top = 12, bottom = 11}})
	obj.btn:SetPoint(self.db.profile.debug.point, self.db.profile.debug.x, self.db.profile.debug.y)
	obj.btn:EnableMouse(true)
	obj.btn:SetMovable(true)
	obj.btn:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	obj.btn:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() FS.db.profile.debug.point, _, _, FS.db.profile.debug.x, FS.db.profile.debug.y = self:GetPoint() end)
	obj.btn.fs = obj.btn:CreateFontString("String", "ARTWORK", "GameFontNormal")
	obj.btn.fs:SetFont(GameFontNormal:GetFont(), 8)
	obj.btn.fs:SetTextColor(1, 1, 1, 1)
	obj.btn.fs:SetJustifyH("LEFT")
	obj.btn.fs:SetJustifyV("TOP")
	for i = 1, 2 do
		local n = "button"..i
		obj.btn[n] = obj.btn:CreateFontString("String", "ARTWORK", "GameFontNormal")
		obj.btn[n]:SetFont(GameFontNormal:GetFont(), 8)
		obj.btn[n]:SetTextColor(1, 1, 1, 1)
		obj.btn[n]:SetJustifyH("LEFT")
		obj.btn[n]:SetJustifyV("TOP")
	end
	obj:GetDebugStr()
	obj:WriteText()
	return obj
end